/* 
 * Copyright (c) 2001 by Dov Grobgeld <dov@imagic.weizmann.ac.il>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <glib.h>
#include <gtk/gtkmain.h>
#include <gtk/gtksignal.h>
#include <gdk-pixbuf/gdk-pixbuf.h>

#include "gtk_image_viewer.h"

#define IMAGE_WIDGET_DEFAULT_SIZE 100
#define DBG(a)

enum {
  VIEW_CHANGED,
  LAST_SIGNAL
};

enum {
  ARG_0,
  ARG_FILE
};

/* Forward declarations */

static void gtk_image_viewer_class_init               (GtkImageViewerClass    *klass);
static void gtk_image_viewer_init                     (GtkImageViewer         *image_viewer);
static void gtk_image_viewer_destroy                  (GtkObject        *object);
static void gtk_image_viewer_realize                  (GtkWidget        *widget);
static void gtk_image_viewer_size_request             (GtkWidget      *widget,
					       GtkRequisition *requisition);
static void gtk_image_viewer_size_allocate            (GtkWidget     *widget,
					       GtkAllocation *allocation);
static gint gtk_image_viewer_expose                   (GtkWidget        *widget,
						GdkEventExpose   *event);
static gint gtk_image_viewer_button_press (GtkWidget      *widget,
					  GdkEventButton *event);
static gint gtk_image_viewer_button_release (GtkWidget        *widget,
					    GdkEventButton   *event);
static gint gtk_image_viewer_motion_notify       (GtkWidget        *widget,
						GdkEventMotion   *event);
static gint gtk_image_viewer_key_press (GtkWidget      *widget,
					GdkEventKey *event);
static gint gtk_image_viewer_enter_notify(GtkWidget *widget,
					  GdkEventCrossing *event);
static gint gtk_image_viewer_leave_notify(GtkWidget *widget,
					  GdkEventCrossing *event);
static gint view_changed(GtkImageViewer *image_widget,
			 int do_force,
			 double scale_x,
			 double scale_y,
			 double x0,
			 double y0);

static gint zoom_reset(GtkWidget *widget);
static gint zoom_in(GtkWidget *widget, int x, int y);
static gint zoom_out(GtkWidget *widget, int x, int y);
static gint zoom_translate(GtkWidget *widget, int dx, int dy);
static void gtk_image_viewer_get_arg(GtkObject    *object,
				     GtkArg       *arg,
				     guint         arg_id);
static void gtk_image_viewer_set_arg(GtkObject    *object,
				     GtkArg       *arg,
				     guint         arg_id);

/* Local data */
static GtkWidgetClass *parent_class = NULL;
static guint gtk_image_viewer_signals[LAST_SIGNAL] = { 0 };

guint
gtk_image_viewer_get_type ()
{
  static guint image_viewer_type = 0;

  if (!image_viewer_type)
    {
      GtkTypeInfo image_viewer_info =
      {
	"GtkImageViewer",
	sizeof (GtkImageViewer),
	sizeof (GtkImageViewerClass),
	(GtkClassInitFunc) gtk_image_viewer_class_init,
	(GtkObjectInitFunc) gtk_image_viewer_init,
	(GtkArgSetFunc) NULL,
	(GtkArgGetFunc) NULL,
      };

      image_viewer_type = gtk_type_unique (gtk_widget_get_type (), &image_viewer_info);
    }

  return image_viewer_type;
}

static void
gtk_image_viewer_class_init (GtkImageViewerClass *class)
{
  GtkObjectClass *object_class;
  GtkWidgetClass *widget_class;

  object_class = (GtkObjectClass*) class;
  widget_class = (GtkWidgetClass*) class;

  parent_class = gtk_type_class (gtk_widget_get_type ());

  object_class->destroy = gtk_image_viewer_destroy;

  widget_class->realize = gtk_image_viewer_realize;
  widget_class->expose_event = gtk_image_viewer_expose;
  widget_class->size_request = gtk_image_viewer_size_request;
  widget_class->size_allocate = gtk_image_viewer_size_allocate;
  widget_class->button_press_event = gtk_image_viewer_button_press;
  widget_class->button_release_event = gtk_image_viewer_button_release;
  widget_class->motion_notify_event = gtk_image_viewer_motion_notify;
  widget_class->key_press_event = gtk_image_viewer_key_press;
  widget_class->enter_notify_event = gtk_image_viewer_enter_notify;
  widget_class->leave_notify_event = gtk_image_viewer_leave_notify;

  gtk_image_viewer_signals[VIEW_CHANGED] =
    gtk_signal_new("view_changed",
		   GTK_RUN_FIRST,
		   object_class->type,
		   GTK_SIGNAL_OFFSET(GtkImageViewerClass, view_changed),
		   gtk_marshal_NONE__NONE,
		   GTK_TYPE_NONE, 0);

  gtk_object_class_add_signals(object_class, gtk_image_viewer_signals, LAST_SIGNAL);

  gtk_object_add_arg_type("GtkImageViewer::file",
			  GTK_TYPE_STRING,
			  GTK_ARG_WRITABLE,
			  ARG_FILE);

  object_class->get_arg = gtk_image_viewer_get_arg;
  object_class->set_arg = gtk_image_viewer_set_arg;
}

/*======================================================================
//  Set default values.
//----------------------------------------------------------------------*/
static void
gtk_image_viewer_init (GtkImageViewer *image_viewer)
{
  GTK_WIDGET_SET_FLAGS (image_viewer, GTK_CAN_FOCUS);
  image_viewer->interp_type = GDK_INTERP_NEAREST;

  image_viewer->do_use_transfer_map = FALSE;
  image_viewer->current_scale_x = 1.0;
  image_viewer->current_scale_y = 1.0;
  image_viewer->current_x0 = 0;
  image_viewer->current_y0 = 0;
  image_viewer->min_zoom = 1.0/8;
  image_viewer->max_zoom = 128;
}

GtkWidget*
gtk_image_viewer_new (GdkPixbuf *_img)
{
  GtkImageViewer *image_viewer;

  image_viewer = gtk_type_new (gtk_image_viewer_get_type ());
  image_viewer->image=NULL;
  
  gtk_image_viewer_set_image(image_viewer, _img);

  return GTK_WIDGET (image_viewer);
}

GtkWidget*
gtk_image_viewer_new_from_file (const char *filename)
{
  GtkWidget *widget;
  GdkPixbuf *image = NULL;

  image = gdk_pixbuf_new_from_file(filename);

  if (image)
    {
      widget = gtk_image_viewer_new(image);
      
      /* The reference count of image should now be == 2. Decrease it. */
      gdk_pixbuf_unref(image);
    }

  return widget;
}

static void
gtk_image_viewer_destroy (GtkObject *object)
{
  GtkImageViewer *image_viewer;

  g_return_if_fail (object != NULL);
  g_return_if_fail (GTK_ISIMAGE_VIEWER (object));

  image_viewer = GTK_IMAGE_VIEWER (object);

  if (image_viewer->image)
    gdk_pixbuf_unref(image_viewer->image);

  if (GTK_OBJECT_CLASS (parent_class)->destroy)
    (* GTK_OBJECT_CLASS (parent_class)->destroy) (object);
}

static void
gtk_image_viewer_realize (GtkWidget *widget)
{
  GtkImageViewer *image_viewer;
  GdkWindowAttr attributes;
  gint attributes_mask;

  g_return_if_fail (widget != NULL);
  g_return_if_fail (GTK_ISIMAGE_VIEWER (widget));

  GTK_WIDGET_SET_FLAGS (widget, GTK_REALIZED);
  image_viewer = GTK_IMAGE_VIEWER (widget);

  attributes.x = widget->allocation.x;
  attributes.y = widget->allocation.y;
  attributes.width = widget->allocation.width;
  attributes.height = widget->allocation.height;
  attributes.wclass = GDK_INPUT_OUTPUT;
  attributes.window_type = GDK_WINDOW_CHILD;
  attributes.event_mask = gtk_widget_get_events (widget) | 
			  GDK_EXPOSURE_MASK |
			  GDK_BUTTON_PRESS_MASK | 
			  GDK_BUTTON_RELEASE_MASK |
			  GDK_POINTER_MOTION_MASK |
                          GDK_POINTER_MOTION_HINT_MASK | 
			  GDK_KEY_PRESS_MASK |
			  GDK_LEAVE_NOTIFY_MASK |
			  GDK_ENTER_NOTIFY_MASK;
  attributes.visual = gtk_widget_get_visual (widget);
  attributes.colormap = gtk_widget_get_colormap (widget);

  attributes_mask = GDK_WA_X | GDK_WA_Y | GDK_WA_VISUAL | GDK_WA_COLORMAP;
  widget->window = gdk_window_new (widget->parent->window, &attributes, attributes_mask);

  widget->style = gtk_style_attach (widget->style, widget->window);

  gdk_window_set_user_data (widget->window, widget);

  gtk_style_set_background (widget->style, widget->window, GTK_STATE_ACTIVE);

}

static void 
gtk_image_viewer_size_request (GtkWidget      *widget,
			       GtkRequisition *requisition)
{
  GtkImageViewer *image_viewer;

  g_return_if_fail (widget != NULL);
  g_return_if_fail (GTK_ISIMAGE_VIEWER (widget));

  image_viewer = GTK_IMAGE_VIEWER (widget);

  if (image_viewer->image)
    {
      requisition->width = gdk_pixbuf_get_width (image_viewer->image);
      requisition->height = gdk_pixbuf_get_height (image_viewer->image);
    }
  else
    {
      requisition->width = IMAGE_WIDGET_DEFAULT_SIZE;
      requisition->height = IMAGE_WIDGET_DEFAULT_SIZE;
    }
}

static void
gtk_image_viewer_size_allocate (GtkWidget     *widget,
			       GtkAllocation *allocation)
{
  GtkImageViewer *image_viewer;
  int old_cnv_w, old_cnv_h;

  g_return_if_fail (widget != NULL);
  g_return_if_fail (GTK_ISIMAGE_VIEWER (widget));
  g_return_if_fail (allocation != NULL);

  widget->allocation = *allocation;
  image_viewer = GTK_IMAGE_VIEWER (widget);

  if (GTK_WIDGET_REALIZED (widget))
    {

      gdk_window_move_resize (widget->window,
			      allocation->x, allocation->y,
			      allocation->width, allocation->height);

      
    }

  old_cnv_w = image_viewer->canvas_width;
  old_cnv_h = image_viewer->canvas_height;
  image_viewer->canvas_width = widget->allocation.width;
  image_viewer->canvas_height = widget->allocation.height;

  /* Update current_x0 and current_y0 if new size is wider than scale *
   * image size!
   */
  if (image_viewer->image)
    {
      double scale_x = image_viewer->current_scale_x;
      double scale_y = image_viewer->current_scale_y;
      int img_w = gdk_pixbuf_get_width(image_viewer->image)*scale_x;
      int img_h = gdk_pixbuf_get_height(image_viewer->image)*scale_y;
      int cnv_w = image_viewer->canvas_width;
      int cnv_h = image_viewer->canvas_height;

      DBG(fprintf(stderr,"Resize: x0 y0 img_w img_h cnv_w cnv_h = %f %f %d %d %d %d\n",
	      image_viewer->current_x0,image_viewer->current_y0,img_w,img_h,cnv_w,cnv_h));

      if (cnv_w > img_w)
	image_viewer->current_x0 = -(cnv_w-img_w)/2;
      else if (old_cnv_w > img_w)
	image_viewer->current_x0 = 0;
      else if (img_w - image_viewer->current_x0 < cnv_w)
	{
	  DBG(fprintf(stderr, "Resize case 3X\n"));
	  image_viewer->current_x0 = img_w-cnv_w;
	}
      if (cnv_h > img_h)
	image_viewer->current_y0 = -(cnv_h-img_h)/2;
      else if (old_cnv_h > img_h)
	image_viewer->current_y0 = 0;
      else if (img_h - image_viewer->current_y0 < cnv_h)
	{
	  DBG(fprintf(stderr, "Resize case 3Y\n"));
	  image_viewer->current_y0 = img_h-cnv_h;
	}
    }
  
}

static gint
gtk_image_viewer_expose (GtkWidget      *widget,
			 GdkEventExpose *event)
{
  GtkImageViewer *image_viewer;

  g_return_val_if_fail (widget != NULL, FALSE);
  g_return_val_if_fail (GTK_ISIMAGE_VIEWER (widget), FALSE);
  g_return_val_if_fail (event != NULL, FALSE);

  image_viewer = GTK_IMAGE_VIEWER (widget);

  gtk_image_viewer_expose_area(image_viewer,
			       event->area.x, event->area.y,
			       event->area.width, event->area.height);
  
  return FALSE;
}

gint
gtk_image_viewer_redraw(GtkWidget *widget)
{
  if (widget->window)
    gtk_image_viewer_expose_area(GTK_IMAGE_VIEWER(widget),
				 0,0, widget->allocation.width, widget->allocation.height);
  return FALSE;
}

static gint
gtk_image_viewer_button_press (GtkWidget      *widget,
			      GdkEventButton *event)
{
  int button = event->button;
  int x = event->x;
  int y = event->y;
  GtkImageViewer *image_viewer;
  image_viewer = GTK_IMAGE_VIEWER (widget);

  if (button == 1)
    zoom_in(widget, x, y);
  else if (button == 2)
    {
      gtk_grab_add (GTK_WIDGET(image_viewer));
      image_viewer->is_mouse_button2_pressed = TRUE;
      image_viewer->last_pan_anchor_x = x;
      image_viewer->last_pan_anchor_y = y;
    }
  else if (button == 3)
    zoom_out(widget, x, y);
  
  image_viewer->button = event->button;
  
  return FALSE;
}

static gint
gtk_image_viewer_button_release (GtkWidget      *widget,
				GdkEventButton *event)
{
  GtkImageViewer *image_viewer;
  int x = event->x;
  int y = event->y;

  g_return_val_if_fail (widget != NULL, FALSE);
  g_return_val_if_fail (GTK_ISIMAGE_VIEWER (widget), FALSE);
  g_return_val_if_fail (event != NULL, FALSE);

  image_viewer = GTK_IMAGE_VIEWER (widget);

  if (image_viewer->button == event->button)
    {
      if (image_viewer->is_mouse_button2_pressed) {
	image_viewer->is_mouse_button2_pressed = FALSE;
	if (image_viewer->last_pan_anchor_x>0 && image_viewer->last_pan_anchor_y > 0)
	    zoom_translate(widget, image_viewer->last_pan_anchor_x-x,
			   image_viewer->last_pan_anchor_y-y);
	gtk_grab_remove (GTK_WIDGET(image_viewer));

	/* Emit an additional view changed signal at the end of the scrolling
	   in order to support redrawing only when the panning is done.
	*/
	gtk_signal_emit(GTK_OBJECT(image_viewer), gtk_image_viewer_signals[VIEW_CHANGED]);
      }
    }

  return FALSE;
}

/*======================================================================
//  The motion callback should scroll the image. The current solution
//  unfortunately does not create smooth scrolling. A further
//  investigation of how e.g. gimp does it is needed.
//----------------------------------------------------------------------*/
static gint
gtk_image_viewer_motion_notify (GtkWidget      *widget,
				GdkEventMotion *event)
{
  GtkImageViewer *image_viewer;
  GdkModifierType state;
  gint x, y;

  g_return_val_if_fail (widget != NULL, FALSE);
  g_return_val_if_fail (GTK_ISIMAGE_VIEWER (widget), FALSE);
  g_return_val_if_fail (event != NULL, FALSE);

  image_viewer = GTK_IMAGE_VIEWER (widget);

  if (event->is_hint) {
     gdk_window_get_pointer(event->window, &x, &y, &state);
  } else {
     x = (int) event->x;
     y = (int) event->y;
  }
  image_viewer->last_x = x;
  image_viewer->last_y = y;

  if (image_viewer->is_mouse_button2_pressed)
    {
      if (image_viewer->last_pan_anchor_x>0 && image_viewer->last_pan_anchor_y > 0)
	zoom_translate(widget, image_viewer->last_pan_anchor_x-x,
		       image_viewer->last_pan_anchor_y-y);
      
      image_viewer->last_pan_anchor_x = x;
      image_viewer->last_pan_anchor_y = y;

      return TRUE;
    }
  
  return FALSE;
}

static gint
gtk_image_viewer_key_press (GtkWidget      *widget,
			    GdkEventKey *event)
{
  GtkImageViewer *image_viewer;
  gint k;

  g_return_val_if_fail (widget != NULL, FALSE);
  g_return_val_if_fail (GTK_ISIMAGE_VIEWER (widget), FALSE);
  g_return_val_if_fail (event != NULL, FALSE);

  image_viewer = GTK_IMAGE_VIEWER (widget);

  k = event->keyval;
    
  if (k == '=' || k == '>')
    {
      zoom_in(widget, image_viewer->last_x, image_viewer->last_y);
    }
  else if (k == '<' || k == '-')
    zoom_out(widget, 0, 0);
  else if (k == '1' || k == 'n')
    zoom_reset(widget);
  return FALSE;
}

static gint gtk_image_viewer_enter_notify(GtkWidget *widget,
					  GdkEventCrossing *event)
{
  if (!GTK_WIDGET_HAS_FOCUS (widget))
    gtk_widget_grab_focus (widget);

  return FALSE;
}

static gint gtk_image_viewer_leave_notify(GtkWidget *widget,
					  GdkEventCrossing *event)
{
  return FALSE;
}

/*======================================================================
//  Accessor related functions.
//----------------------------------------------------------------------*/
guint
gtk_image_viewer_set_image(GtkImageViewer *widget,
			   GdkPixbuf *_pixbuf)
{
  GtkImageViewer *image_viewer;

  g_return_val_if_fail (widget != NULL, FALSE);
  g_return_val_if_fail (GTK_ISIMAGE_VIEWER (widget), FALSE);

  image_viewer = GTK_IMAGE_VIEWER (widget);

  if (widget->image)
    gdk_pixbuf_unref(widget->image);

  if (_pixbuf)
    image_viewer->image = gdk_pixbuf_ref(_pixbuf);
  else
    image_viewer->image = NULL;

  gtk_image_viewer_redraw(GTK_WIDGET(widget));
  
  return 0;
}

guint
gtk_image_viewer_set_file(GtkImageViewer *widget,
			  char *filename)
{
  GtkImageViewer *image_viewer;
  GdkPixbuf *image = NULL;

  image = gdk_pixbuf_new_from_file(filename);
  gtk_image_viewer_set_image(widget, image);
  if (image)
    gdk_pixbuf_unref(image);

  return 0;
}

guint
gtk_image_viewer_set_zoom_range(GtkImageViewer *widget,
				double min, double max)
{
  GtkImageViewer *image_viewer;

  g_return_val_if_fail (widget != NULL, FALSE);
  g_return_val_if_fail (GTK_ISIMAGE_VIEWER (widget), FALSE);

  image_viewer = GTK_IMAGE_VIEWER (widget);

  image_viewer->min_zoom = min;
  image_viewer->max_zoom = max;

  return 1;
}

gboolean
gtk_image_viewer_get_is_panning(GtkImageViewer *widget)
{
  return widget->is_mouse_button2_pressed;
}

guint
gtk_image_viewer_get_canvas_width(GtkImageViewer *widget)
{
  return widget->canvas_width;
}

guint
gtk_image_viewer_get_canvas_height(GtkImageViewer *widget)
{
  return widget->canvas_height;
}

void
gtk_image_viewer_get_scale(GtkImageViewer *widget,
			   /* output */
			   double *scale_x,
			   double *scale_y
			   )
{
  *scale_x = widget->current_scale_x;
  *scale_y = widget->current_scale_y;
}

GdkPixbuf *
gtk_image_viewer_get_image(GtkImageViewer *widget)
{
  return widget->image;
}

/*======================================================================
// gtk object system support...
//----------------------------------------------------------------------*/
static void
gtk_image_viewer_set_arg(GtkObject    *object,
                        GtkArg       *arg,
                        guint         arg_id)
{
  GtkImageViewer *image_viewer;

  image_viewer = GTK_IMAGE_VIEWER(object);

  switch(arg_id)
  {
  case ARG_FILE:
    gtk_image_viewer_set_file(image_viewer, GTK_VALUE_STRING(*arg));
    break;
  default:
    break;
  }
}

static void
gtk_image_viewer_get_arg(GtkObject    *object,
                        GtkArg       *arg,
                        guint         arg_id)
{
  GtkImageViewer *image_viewer;

  image_viewer = GTK_IMAGE_VIEWER(object);

  switch(arg_id)
  {
  default:
    arg->type = GTK_TYPE_INVALID;
    break;
  }
}

/*======================================================================
//  Zoom related functions.
//----------------------------------------------------------------------*/
static gint
zoom_reset(GtkWidget *widget)
{
    view_changed(GTK_IMAGE_VIEWER(widget), TRUE, 1, 1, 0, 0);
    return 1;
}

static gint
zoom_in(GtkWidget *widget, int x, int y)
{
  GtkImageViewer *image_viewer = GTK_IMAGE_VIEWER (widget);
  
  gtk_image_viewer_zoom_around_fixed_point(image_viewer,
					   image_viewer->current_scale_x * 2,
					   image_viewer->current_scale_y * 2,
					   x, y,
					   image_viewer->canvas_width/2,
					   image_viewer->canvas_height/2);

  return 1;
}

static gint
zoom_out(GtkWidget *widget, int x, int y)
{
  GtkImageViewer *image_viewer = GTK_IMAGE_VIEWER (widget);

  gtk_image_viewer_zoom_around_fixed_point(image_viewer,
					   image_viewer->current_scale_x/2,
					   image_viewer->current_scale_y/2,
					   image_viewer->canvas_width/2,
					   image_viewer->canvas_height/2,
					   image_viewer->canvas_width/2,
					   image_viewer->canvas_height/2);
  return 1;
}

static gint
zoom_translate(GtkWidget *widget, int dx, int dy)
{
  GtkImageViewer *image_viewer = GTK_IMAGE_VIEWER (widget);
  
  view_changed(image_viewer,
	       FALSE,
	       image_viewer->current_scale_x,
	       image_viewer->current_scale_y,
	       image_viewer->current_x0+dx,
	       image_viewer->current_y0+dy);
  return 1;
}

/*======================================================================
//  Expose an area. x0 y0 x1 and y1 are in pixel coordinates.
//----------------------------------------------------------------------*/
void
gtk_image_viewer_expose_area(GtkImageViewer *image_viewer,
			     int exp_x0, int exp_y0,
			     int w, int h)
{
  GtkWidget *widget = GTK_WIDGET(image_viewer); /* In order to access window */
  GdkPixbuf *img_scaled;
  double scale_x = image_viewer->current_scale_x;
  double scale_y = image_viewer->current_scale_y;
  int exp_x1 = exp_x0 + w;
  int exp_y1 = exp_y0 + h;

  DBG(fprintf(stderr, "expose: %d %d %d %d\n", exp_x0, exp_y0, w, h));
  if (image_viewer->image)
    {
				/* Canvas size */
      int cnv_w = image_viewer->canvas_width;
      int cnv_h = image_viewer->canvas_height;
      int copy_w, copy_h;
      int img_w, img_h;		/* Image's size (pixels) */
      int offs_x, offs_y;
      int dst_x, dst_y;
				/* Canvas coordinates in which drawing
				   is done. */
      int img_x0 = 0;
      int img_x1 = cnv_w;
      int img_y0 = 0;
      int img_y1 = cnv_h;
      dst_x = dst_y = 0;
      copy_w = w;
      copy_h = h;
      img_w = gdk_pixbuf_get_width(image_viewer->image);
      img_h = gdk_pixbuf_get_height(image_viewer->image);
      offs_x = -image_viewer->current_x0-exp_x0;
      offs_y = -image_viewer->current_y0-exp_y0;

      /* If img fits in canvas horizontally, need a smaller drawing zone */
      if (img_w * scale_x < cnv_w) 
	{
	  img_x0 = (cnv_w - img_w*scale_x)/2;
	  img_x1 = img_x0 + img_w*scale_x;
	}

      /* If img fits in canvas vertically : need a smaller drawing zone */
      if (img_h * scale_y < cnv_h)
	{
	  img_y0 = (cnv_h - img_h*scale_y)/2;
	  img_y1 = img_y0 + img_h*scale_y;
	}

      if (exp_x1 < img_x0)
	{
	  DBG(fprintf(stderr, "Case 1X\n"));
	  copy_w = 0;
	  offs_x = 0;
	  dst_x = img_x0;
	  gdk_draw_rectangle(widget->window,
			     widget->style->bg_gc[GTK_WIDGET_STATE (widget)],
			     TRUE,
			     exp_x0,exp_y0,w, h);
	}
      else if (exp_x0 < img_x0 && exp_x1 > img_x1)
	{
	  DBG(fprintf(stderr, "Case 2X\n"));
	  copy_w = w - (img_x0 - exp_x0) - (exp_x1 - img_x1);
	  dst_x = img_x0;
	  offs_x = 0;
	  
	  gdk_draw_rectangle(widget->window,
			     widget->style->bg_gc[GTK_WIDGET_STATE (widget)],
			     TRUE,
			     exp_x0,exp_y0,(img_x0-exp_x0), h);
	  gdk_draw_rectangle(widget->window,
			     widget->style->bg_gc[GTK_WIDGET_STATE (widget)],
			     TRUE,
			     img_x1,exp_y0,(exp_x1-img_x1), h);
	}
      else if (exp_x0 < img_x0)
	{
	  DBG(fprintf(stderr, "Case 3X\n"));
	  offs_x = 0;
	  copy_w = w - (img_x0 - exp_x0);
	  dst_x = img_x0;
	  
	  gdk_draw_rectangle(widget->window,
			     widget->style->bg_gc[GTK_WIDGET_STATE (widget)],
			     TRUE,
			     exp_x0,exp_y0,(img_x0-exp_x0), h);
	}
      else if (exp_x0 > img_x1)
	{
	  DBG(fprintf(stderr, "Case 4X\n"));
	  copy_w = 0;
	  gdk_draw_rectangle(widget->window,
			     widget->style->bg_gc[GTK_WIDGET_STATE (widget)],
			     TRUE,
			     exp_x1,exp_y0,w, h);
	  
	}
      else if (exp_x1 > img_x1)
	{
	  DBG(fprintf(stderr, "Case 5X\n"));
	  copy_w = w - (exp_x1 - img_x1);
	  dst_x = exp_x0;
	  offs_x = -(exp_x0-img_x0);
	  gdk_draw_rectangle(widget->window,
			     widget->style->bg_gc[GTK_WIDGET_STATE (widget)],
			     TRUE,
			     exp_x1,exp_y0,exp_x1-img_x1, h);
	}
      else
	{
	  DBG(fprintf(stderr, "Case 6X\n"));
	  dst_x = exp_x0;
	  copy_w = w;
	}

      /* Y dir */
      if (exp_y1 < img_y0)
	{
	  DBG(fprintf(stderr, "Case 1Y\n"));
	  copy_h = 0;
	  offs_y = 0;
	  dst_x = img_y0;
	  gdk_draw_rectangle(widget->window,
			     widget->style->bg_gc[GTK_WIDGET_STATE (widget)],
			     TRUE,
			     exp_x0,exp_y0,w, h);
	}
      else if (exp_y0 < img_y0 && exp_y1 > img_y1)
	{
	  DBG(fprintf(stderr, "Case 2Y\n"));
	  copy_h = h - (img_y0 - exp_y0) - (exp_y1 - img_y1);
	  dst_y = img_y0;
	  offs_y = 0;
	  
	  gdk_draw_rectangle(widget->window,
			     widget->style->bg_gc[GTK_WIDGET_STATE (widget)],
			     TRUE,
			     exp_x0,exp_y0,w, img_y0-exp_y0);
	  gdk_draw_rectangle(widget->window,
			     widget->style->bg_gc[GTK_WIDGET_STATE (widget)],
			     TRUE,
			     exp_x0, img_y1,w, exp_y1-img_y1);
	}
      else if (exp_y0 < img_y0)
	{
	  DBG(fprintf(stderr, "Case 3Y\n"));
	  offs_y = 0;
	  copy_h = h - (img_y0 - exp_y0);
	  dst_y = img_y0;
	  
	  gdk_draw_rectangle(widget->window,
			     widget->style->bg_gc[GTK_WIDGET_STATE (widget)],
			     TRUE,
			     exp_x0,exp_y0,w,(img_y0-exp_y0));
	}
      else if (exp_y0 > img_y1)
	{
	  DBG(fprintf(stderr, "Case 4Y\n"));
	  copy_h = 0;
	  gdk_draw_rectangle(widget->window,
			     widget->style->bg_gc[GTK_WIDGET_STATE (widget)],
			     TRUE,
			     exp_x0,exp_y1,w, h);
	  
	}
      else if (exp_y1 > img_y1)
	{
	  DBG(fprintf(stderr, "Case 5Y\n"));
	  copy_h = h - (exp_y1 - img_y1);
	  dst_y = exp_y0;
	  offs_y = -(exp_y0-img_y0);
	  DBG(fprintf(stderr, "copy_h dst_y = %d %d\n", copy_h, dst_y));
	  gdk_draw_rectangle(widget->window,
			     widget->style->bg_gc[GTK_WIDGET_STATE (widget)],
			     TRUE,
			     exp_x0,exp_y1,w,exp_y1-img_y1);
	}
      else
	{
	  DBG(fprintf(stderr, "Case 6Y\n"));
	  dst_y = exp_y0;
	  copy_h = h;
	}
      

      DBG(fprintf(stderr, "dst_x dst_y copy_w copy_h offs_x offs_y = %d %d %d %d %f %f\n",
	      dst_x, dst_y, copy_w, copy_h,
	      image_viewer->current_x0, image_viewer->current_y0));
      /* Scale and copy the image */
      if (copy_w > 0 && copy_h > 0)
	{
	  img_scaled = gdk_pixbuf_new(gdk_pixbuf_get_colorspace(image_viewer->image),
				      FALSE,
				      gdk_pixbuf_get_bits_per_sample(image_viewer->image),
				      copy_w, copy_h);

	  /* Check legality */
	  DBG(fprintf(stderr, "offs_y copy_h scale_y h = %d %d %f %d\n",
		      offs_y, copy_h, scale_y, img_h));
	  gdk_pixbuf_scale(image_viewer->image,
			   img_scaled,
			   0,0,
			   copy_w,
			   copy_h,
			   offs_x,
			   offs_y,
			   scale_x, scale_y,
			   image_viewer->interp_type);
	  
	  if (image_viewer->do_use_transfer_map)
	    {
	      
	      /* Assume we are doing 8-bit... */
	      guint8 *buf_scaled = gdk_pixbuf_get_pixels(img_scaled);
	      gint w = gdk_pixbuf_get_width(img_scaled);
	      gint h = gdk_pixbuf_get_height(img_scaled);
	      gint rs = gdk_pixbuf_get_rowstride(img_scaled);
	      gint row_idx, col_idx;
	      guint8 *rmap = image_viewer->transfer_map[0];
	      guint8 *gmap = image_viewer->transfer_map[1];
	      guint8 *bmap = image_viewer->transfer_map[2];
	      
	      for (row_idx=0; row_idx<h; row_idx++)
		{
		  guint8 *p = &buf_scaled[rs*row_idx];
		  
		  for (col_idx=0; col_idx<w; col_idx++)
		    {
		      *p++ = rmap[*p];
		      *p++ = gmap[*p];
		      *p++ = bmap[*p];
		    }
		}
	    }

	  gdk_pixbuf_render_to_drawable (img_scaled,
					 widget->window,
					 widget->style->white_gc,
					 0, 0,
					 dst_x, dst_y,
					 copy_w, copy_h,
					 GDK_RGB_DITHER_NORMAL,
					 image_viewer->current_x0,
					 image_viewer->current_y0);
	  
	  gdk_pixbuf_unref(img_scaled);
	}
    }
  else
    {
      gdk_draw_rectangle(widget->window,
			 widget->style->bg_gc[GTK_WIDGET_STATE (widget)],
			 TRUE,
			 exp_x0,exp_y0,w, h);
    }

  gtk_signal_emit(GTK_OBJECT(image_viewer), gtk_image_viewer_signals[VIEW_CHANGED]);

}


/*======================================================================
//  gtk_image_viewer_zoom_around_fixed_point is a preprocessor
//  to view_changed that allows a zooming while moving an old pixel
//  coordinate to a given new position. E.g. moving the x,y coordinate
//  clicked by the mouse to the center of the zoomed up image.
//----------------------------------------------------------------------*/
gint
gtk_image_viewer_zoom_around_fixed_point(GtkImageViewer *image_viewer,
					 double new_scale_x,
					 double new_scale_y,
					 double old_x,
					 double old_y,
					 double new_x,
					 double new_y)
{
  double old_scale_x, old_scale_y, old_x0, old_y0, new_x0, new_y0;

  g_return_val_if_fail (image_viewer != NULL, FALSE);
  g_return_val_if_fail (GTK_ISIMAGE_VIEWER (image_viewer), FALSE);

  /* Short cuts */
  old_x0 = image_viewer->current_x0;
  old_y0 = image_viewer->current_y0;
  old_scale_x = image_viewer->current_scale_x;
  old_scale_y = image_viewer->current_scale_y;

  new_x0 = new_scale_x/old_scale_x * (old_x + old_x0) - new_x;
  new_y0 = new_scale_y/old_scale_y * (old_y + old_y0) - new_y;
  
  view_changed(image_viewer, FALSE, new_scale_x, new_scale_y, new_x0, new_y0);

  return 0;
}


/*======================================================================
//  Make sure graphics expose events are processed before scrolling
//  again. This doesn't really work. Does anyone know how to get
//  rid of the exposure delays?
//----------------------------------------------------------------------*/
static void 
process_exposes (GtkWidget *widget)
{
  GdkEvent *event;
  
  while ((event = gdk_event_get_graphics_expose (widget->window)) != NULL)
    {
      gtk_widget_event (widget, event);
      if (event->expose.count == 0)
	{
	  gdk_event_free (event);
	  break;
	}
      gdk_event_free (event);
    }
}

/*======================================================================
//  view_changed does clipping, scrolling and scaling.
//----------------------------------------------------------------------*/
static gint
view_changed(GtkImageViewer *image_viewer,
	     int do_force, double scale_x, double scale_y, double x0, double y0)
{
  GtkWidget *widget;
  GdkPixbuf *im;
  int render_width, render_height;

  g_return_val_if_fail (image_viewer != NULL, FALSE);
  g_return_val_if_fail (GTK_ISIMAGE_VIEWER (image_viewer), FALSE);

  widget = GTK_WIDGET(image_viewer);

  im = image_viewer->image; /* A short cut */

  /* This should certainly be configurable */
  if (scale_x < 1.0)
    image_viewer->interp_type = GDK_INTERP_BILINEAR;
  else
    image_viewer->interp_type = GDK_INTERP_NEAREST;

  /* Clip the scale */
  if (scale_x>image_viewer->max_zoom && scale_y > image_viewer->max_zoom)
    return 0;
  else if (scale_x < image_viewer->min_zoom && scale_y < image_viewer->min_zoom)
    return 0;

  /* Clip the request */
  if (im)
    {
      int width = gdk_pixbuf_get_width(im);
      int height = gdk_pixbuf_get_height(im);
      int cwidth = image_viewer->canvas_width;
      int cheight = image_viewer->canvas_height;
      
      if (width*scale_x > cwidth)
	  render_width = cwidth;
      else
	  render_width = width * scale_x;
      
      if (height*scale_y > cheight)
	  render_height = cheight;
      else
	  render_height = height * scale_y;

      if (render_width < cwidth)
	x0 = -(cwidth - render_width)/2;
      else if (x0 + render_width > width*scale_x)
	x0 = width*scale_x - render_width;
      else if (x0<0)
	x0 = 0;

      if (render_height < cheight)
	y0 = -(cheight - render_height)/2;
      else if (y0 + render_height > height*scale_y)
	y0 = height*scale_y - render_height;
      else if (y0<0)
	y0 = 0;
    }


  /* If scale is the same, then the image has only been scrolled,
     and we only need to update the exposed areas.
   */
  if (widget->window && scale_x == image_viewer->current_scale_x
      && widget->window && scale_y == image_viewer->current_scale_y
      )
    {
      int dx = (x0 - image_viewer->current_x0);
      int dy = (y0 - image_viewer->current_y0);
      int src_x = (dx < 0) ? 0 : dx;
      int src_y = (dy < 0) ? 0 : dy;
      int dst_x = (dx < 0) ? -dx : 0;
      int dst_y = (dy < 0) ? -dy : 0;
      GdkGC *gc = gdk_gc_new(widget->window); /* Should be static in widget */
      
      gdk_gc_copy(gc, widget->style->white_gc);
      gdk_gc_set_exposures(gc, TRUE);
      
      image_viewer->current_x0 = x0;
      image_viewer->current_y0 = y0;

      /* Scroll visible region */
      gdk_draw_pixmap (widget->window,
		       gc,
		       widget->window,
		       src_x, src_y,
		       dst_x, dst_y,
		       image_viewer->canvas_width - abs (dx),
		       image_viewer->canvas_height - abs (dy));

      process_exposes(widget);
      gdk_gc_unref(gc);

      /* And fill in the new areas */
      if (dx)
	{
	  src_x = (dx < 0) ? 0 : image_viewer->canvas_width - dx;
	  src_y = 0;
	  gtk_image_viewer_expose_area (image_viewer,
					src_x, src_y,
					abs (dx), image_viewer->canvas_height);
	}
      if (dy)
	{
	  src_x = 0;
	  src_y = (dy < 0) ? 0 : image_viewer->canvas_height - dy;
	  gtk_image_viewer_expose_area (image_viewer,
					src_x, src_y,
					image_viewer->canvas_width, abs (dy));
	}
      
    }
  
  else if (do_force
      || scale_x != image_viewer->current_scale_x
      || scale_y != image_viewer->current_scale_y
      || x0 != image_viewer->current_x0
      || y0 != image_viewer->current_y0)
    {

      /* Remember the current transform */
      image_viewer->current_scale_x = scale_x;
      image_viewer->current_scale_y = scale_y;
      image_viewer->current_x0 = x0;
      image_viewer->current_y0 = y0;
      gtk_image_viewer_redraw(widget);
    }
  
  return 1;
}

/*======================================================================
//  Utility functions for converting between the image coordinates
//  and the canvas coordinate.
//----------------------------------------------------------------------*/
void
gtk_image_viewer_img_coord_to_canv_coord(GtkImageViewer *image_viewer,
					 double imgx, double imgy,
					 /* output */
					 double *cx, double *cy)
{
  double x0 = image_viewer->current_x0;
  double y0 = image_viewer->current_y0;
  double sx = image_viewer->current_scale_x;
  double sy = image_viewer->current_scale_y;
  *cx = imgx * sx - x0;
  *cy = imgy * sy - y0;
}

void
gtk_image_viewer_canv_coord_to_img_coord(GtkImageViewer *image_viewer,
					 double cx, double cy,
					 /* output */
					 double *imgx, double *imgy)
{
  double x0 = image_viewer->current_x0;
  double y0 = image_viewer->current_y0;
  double sx = image_viewer->current_scale_x;
  double sy = image_viewer->current_scale_y;

  *imgx  = (cx + x0) / sx;
  *imgy  = (cy + y0) / sy;
}

/*======================================================================
//  Set the transfer function for the display. Note that the widget
//  does not own or replicate the memory of the maps.
//----------------------------------------------------------------------*/
void
gtk_image_viewer_set_transfer_map(GtkImageViewer *widget,
				  guint8 *rmap,
				  guint8 *gmap,
				  guint8 *bmap
				  )
{
  GtkImageViewer *image_viewer = GTK_IMAGE_VIEWER(widget);

  g_return_if_fail (widget != NULL);
  g_return_if_fail (GTK_ISIMAGE_VIEWER (widget));
  
  image_viewer->transfer_map[0] = rmap;
  image_viewer->transfer_map[1] = gmap;
  image_viewer->transfer_map[2] = bmap;
  image_viewer->do_use_transfer_map = TRUE;

  gtk_image_viewer_redraw(GTK_WIDGET(widget));

}

void           gtk_image_viewer_get_transform(GtkImageViewer *widget,
					      /* output */
					      double *x,
					      double *y,
					      double *scale_x,
					      double *scale_y)
{
    *x = widget->current_x0;
    *y = widget->current_y0;
    *scale_x = widget->current_scale_x;
    *scale_y = widget->current_scale_y;
}

void           gtk_image_viewer_set_transform(GtkImageViewer *widget,
					      double x,
					      double y,
					      double scale_x,
					      double scale_y)
{
    view_changed(widget, FALSE, scale_x, scale_y, x,y);
}
